
/******************************************************************\
*                                                                  *
*                        Buffer mapy                               *
*                                                                  *
\******************************************************************/

#include "stdafx.h"
#include "peter.h"
#include "MainFrm.h"
#include "IconBuf.h"
#include "TextBuf.h"
#include "RealBuf.h"
#include "BoolBuf.h"
#include "MapBuf.h"
#include "SprtBuf.h"
#include "SndBuf.h"
#include "ProgBuf.h"
#include "EditDial.h"
#include "EditSprt.h"
#include "EditIcon.h"
#include "EditMap.h"
#include "EditSnd.h"
#include "Prog.h"
#include "Exec.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define DATAREZERVA		0x1000	// rezerva pi zvtovn bufferu dat (bajt)
#define INDEXREZERVA	100		// rezerva pi zvtovn bufferu index (poloek)

////////////////////////////////////////////////////////////////////
// konstruktor a destruktor

CMapBuf::CMapBuf()
{
	m_Data = NULL;			// nejsou data mapy
	m_Index = NULL;			// nejsou indexy mapy
	m_Num = 0;				// nen dn poloka v bufferu
	m_DataNum = 0;			// poet bajt v bufferu dat mapy
	m_DataMax = 0;			// velikost bufferu dat mapy
	m_IndexMax = 0;			// velikost bufferu index map (poloek)
	m_ProgBuf = NULL;		// nen buffer programu
}

CMapBuf::~CMapBuf()
{
	Aloc(m_Data,0);			// zruen bufferu dat
	Aloc(m_Index,0);		// zruen bufferu index
}


////////////////////////////////////////////////////////////////////
// pidn poloky do bufferu (vrac index)

int CMapBuf::Add(MAP* map)
{
// velikost pidvan poloky v bajtech
	int size = MapSize(map);

// zvten bufferu index
	if (m_Num >= m_IndexMax)
	{
		m_IndexMax = m_Num + INDEXREZERVA;
		Aloc(m_Index, m_IndexMax);
	}

// zvten bufferu dat
	if (m_DataNum+size > m_DataMax)
	{
		m_DataMax = m_DataNum + size + DATAREZERVA;
		Aloc(m_Data, m_DataMax);
	}

// kopie dat poloky
	memcpy(m_Data + m_DataNum, map, size);

// nastaven indexu na poloku
	m_Index[m_Num] = m_DataNum;

// zven tae poloek
	m_Num++;

// zven ukazatele velikosti dat v bufferu
	m_DataNum += size;

// navrcen indexu nov poloky
	return m_Num-1;
}


////////////////////////////////////////////////////////////////////
// vytvoen nov mapy danch rozmr (vrac index)

int CMapBuf::New(int width, int height)
{
// omezen velikosti mapy
	if (width > MAXMAPWIDTH)
		width = MAXMAPWIDTH;

	if (height > MAXMAPHEIGHT)
		height = MAXMAPHEIGHT;

	if (width < 1)
		width = 1;

	if (height < 1)
		height = 1;

	if (width * height > MAXMAPSIZE)
		height = MAXMAPSIZE / width;

// velikost nov poloky v bajtech
	int size = (width*height-1)*sizeof(MAPITEM) + sizeof(MAP);

// zvten bufferu index
	if (m_Num >= m_IndexMax)
	{
		m_IndexMax = m_Num + INDEXREZERVA;
		Aloc(m_Index, m_IndexMax);
	}

// zvten bufferu dat
	if (m_DataNum+size > m_DataMax)
	{
		m_DataMax = m_DataNum + size + DATAREZERVA;
		Aloc(m_Data, m_DataMax);
	}

// adresa nov poloky
	MAP* map = (MAP*)(m_Data + m_DataNum);

// inicializace dat poloky
	map->width = width;
	map->height = height;
	memset(&map->item[0], 0, width*height*sizeof(MAPITEM));

// nastaven indexu na poloku
	m_Index[m_Num] = m_DataNum;

// zven tae poloek
	m_Num++;

// zven ukazatele velikosti dat v bufferu
	m_DataNum += size;

// navrcen indexu nov poloky
	return m_Num-1;
}


////////////////////////////////////////////////////////////////////
// nastaven nov velikosti mapy

void CMapBuf::SetSize(int Index, int width, int height)
{
// lokln promnn
	int size;							// nov velikost poloky
	int oldwidth;						// pvodn ka poloky
	int oldheight;						// pvodn vka poloky
	int oldsize;						// pvodn velikost poloky
	int difsize;						// rozdl velikost (zven)
	int i;								// pracovn ta
	BYTE* src;							// ukazatele pi operacch pesunu
	BYTE* dst;

// kontrola indexu
	if ((Index < 0) || (Index >= m_Num))
		return;

// omezen velikosti mapy
	if (width > MAXMAPWIDTH)
		width = MAXMAPWIDTH;

	if (height > MAXMAPHEIGHT)
		height = MAXMAPHEIGHT;

	if (width < 1)
		width = 1;

	if (height < 1)
		height = 1;

	if (width * height > MAXMAPSIZE)
		height = MAXMAPSIZE / width;

// pvodn velikost poloky v bajtech
	oldsize = MapSize(Index);			// pvodn velikost

// nov velikost poloky v bajtech, rozdl velikost
	size = (width*height-1)*sizeof(MAPITEM) + sizeof(MAP);
	difsize = size - oldsize;			// rozdl velikost

// zvten bufferu dat, je-li poteba
	if (m_DataNum + difsize > m_DataMax)
	{
		m_DataMax = m_DataNum + difsize + DATAREZERVA;
		Aloc(m_Data, m_DataMax);
	}

// adresa korigovan poloky
	MAP* map = Get(Index);

// schova pvodnch parametr poloky
	oldwidth = map->width;				// pvodn ka poloky
	oldheight = map->height;			// pvodn vka poloky

// oprava potu dk
	if (height != oldheight)
	{
		src = (BYTE*)map + (oldwidth*oldheight-1)*sizeof(MAPITEM) + sizeof(MAP);
		dst = (BYTE*)map + (oldwidth*height-1)*sizeof(MAPITEM) + sizeof(MAP);

		if (Index < m_Num-1)
		{
			memmove(dst, src, m_DataNum - m_Index[Index+1]);
		}
		m_DataNum += (height-oldheight)*oldwidth*sizeof(MAPITEM);
		map->height = height;				// nov vka poloky

		if (height > oldheight)
		{
			memset(src, 0, (height - oldheight)*oldwidth*sizeof(MAPITEM));
		}
	}

// oprava dlky dk
	if (width != oldwidth)
	{
		src = (BYTE*)map + (oldwidth - 1)*sizeof(MAPITEM) + sizeof(MAP);
		dst = (BYTE*)map + (width - 1)*sizeof(MAPITEM) + sizeof(MAP);

		for (i = 0; i < height; i++)
		{
			memmove(dst, src, m_DataNum - (m_Index[Index] + (i*width+oldwidth-1)*sizeof(MAPITEM) + sizeof(MAP)));

			if (width > oldwidth)
			{
				memset(src, 0, (width-oldwidth)*sizeof(MAPITEM));
			}

			src += width*sizeof(MAPITEM);
			dst += width*sizeof(MAPITEM);
			m_DataNum += (width - oldwidth)*sizeof(MAPITEM);
		}
		map->width = width;					// nov ka poloky
	}

// korekce index na ostatn poloky 
	for (i = Index+1; i < m_Num; i++)
	{
		m_Index[i] += difsize;
	}
}


////////////////////////////////////////////////////////////////////
// zruen poloky

void CMapBuf::Delete(int Index)
{
// lokln promnn
	int i, b;					// pracovn tae
	int	size;					// velikost ruench dat
	CProgBuf* buf;				// ukazatel na programov buffer
	PROGITEM* item;				// ukazatel na poloku programu

// kontrola indexu
	if ((Index < 0) || (Index >= m_Num))
		return;

// poet bajt ke zruen
	size = MapSize(Index);

// zruen dat poloky (psun zbylch poloek)
	if (Index < m_Num-1)
		memmove(&(m_Data[m_Index[Index]]), &(m_Data[m_Index[Index+1]]), 
			m_DataNum - m_Index[Index+1]);

// snen tae dat v bufferu
	m_DataNum -= size;

// zruen indexu a korekce zbylch index
	for (i = Index; i < m_Num-1; i++)
	{
		m_Index[i] = m_Index[i+1] - size;
	}

// snen tae poloek
	m_Num--;

// korekce referenc v programovch bufferech
	if (m_ProgBuf != NULL)
	{
		for (b = 0; b < PROGBUFNUM; b++) // ukazatel sla bufferu
		{
			buf = &m_ProgBuf[b]; // adresa bufferu

			for (i = buf->m_Num-1; i >= 0; i--) // tn indexu poloky dol
			{
				item = &(buf->m_Items[i]);		// schova adresy poloky

				if ((item->data == BufMapID) &&	// odkaz na buffer mapy
					(item->index >= Index))		// index, kter bude ovlivnn
				{
					if (item->index == Index)	// je to ruen index ?
					{
						item->index = -1;		// zruen odkazu
						item->data = -1;
					}
					else
						item->index--;			// jinak korekce indexu
				}
			}
		}
	}
}


////////////////////////////////////////////////////////////////////
// duplikace poloky

int CMapBuf::Duplik(int Index)
{
// korekce indexu
	if ((Index < 0) || (Index >= m_Num))
		Index = 0;

// velikost pidvan poloky v bajtech
	int size = MapSize(Index);

// zvten bufferu index
	if (m_Num >= m_IndexMax)
	{
		m_IndexMax = m_Num + INDEXREZERVA;
		Aloc(m_Index, m_IndexMax);
	}

// zvten bufferu dat
	if (m_DataNum+size > m_DataMax)
	{
		m_DataMax = m_DataNum + size + DATAREZERVA;
		Aloc(m_Data, m_DataMax);
	}

// kopie dat poloky
	memcpy(m_Data + m_DataNum, Get(Index), size);

// nastaven indexu na poloku
	m_Index[m_Num] = m_DataNum;

// zven tae poloek
	m_Num++;

// zven ukazatele velikosti dat v bufferu
	m_DataNum += size;

// navrcen indexu nov poloky
	return m_Num-1;
}

